"""Base class and utility functions for Executors."""

import datetime
import logging
import os
import re
from abc import ABCMeta, abstractmethod

from pathos.multiprocessing import Pool

logger = logging.getLogger(__name__)
# set default log level as debug
# logger.setLevel(logging.DEBUG)


class BaseExecutor(object):
    """Base class for a task executor.

    An Executor is responsible for running tasks that are provided to it
    by an Optimizer.

    An *Executor subclass must implement a run method that accepts a list of
    commands to run in some environment (e.g. locally, on a job scheduler).
    The run method will handle parallel or sequential marshalling of the tasks.
    The run method must call the BaseExecutor.wrap_cmd method which implements
    a trace file of command state and output, plus tracking of optimization
    scores. BaseExecutor.wrap_cmd calls back to the _run_cmd method of the
    Executor subclass which must implement the low-level execution of a single
    executable command.

    """

    metaclass = ABCMeta

    conc_tasks = 1
    """Number of concurrent tasks the executor will run.
    Computed on __init__ in the subclass"""

    logdir = ""
    params = None

    def __init__(self, cwd, params):
        """Initialize the Executor, work and log directories."""
        logger.info("Initializing %s " % self.__class__.__name__)

        self.executor = 'base'

        self.params = params

        if 'work_dir' in params.vals and os.path.isdir(params.vals['work_dir']):
            cwd = params.vals['work_dir']

        if not os.path.isdir(cwd):
            cwd = '.'

        logger.info("Changed to work directory: %s" % cwd)
        os.chdir(cwd)

        logdir = os.path.join(cwd,
                              "param_runner_%s" % datetime.datetime.now().strftime(
                                  '%Y%m%d-%H%M%S'))

        try:
            os.mkdir(logdir)
            logger.info(
                " - Trace file and output will be at: %s" % os.path.relpath(
                    logdir))
        except IOError:
            raise IOError("Could not create log directory: %s" % logdir)

        self.logdir = logdir

        self.start_trace()

    def get_pool(self):
        """Return a process pool an optimizer can use to run tasks in parallel.

        The number of tasks that can run in parallel depends on the setup of the
        executor, so the optimizer must retrieve the pool, and not manage its own.
        """
        return Pool(self.conc_tasks)

    def start_trace(self):
        """Write out the header line of the trace.txt TSV file."""
        trace_path = os.path.join(self.logdir, 'trace.txt')
        trace_file = open(trace_path, 'w')

        trace_file.write("Index\tStatus\t")

        if 'optimizer' in self.params.vals.keys() and self.params.vals["optimizer"] in ["spearmint", "ray_tune"]:
            trace_file.write("Score\t")
        else:
            for param in self.params.vals['parameters']:
                trace_file.write("%s\t" % param['id'])

        trace_file.write("Command\tReturn Code\tDuration\tStart\tEnd")

        if "summary" in self.params.vals:
            for summary in self.params.vals['summary']:
                trace_file.write("\t%s" % summary['id'])

        trace_file.write("\n")

    def write_trace(self, summary_dict):
        """Write a single line to the trace file, summarizing a task."""
        print('Start to writing trace ...')
        logger.debug(summary_dict)

        try:
            self._write_trace(summary_dict)
        except Exception as e:
            if isinstance(summary_dict, dict) and ("_task_index" in summary_dict):
                logger.error(
                    "[%s] writing trace for task %d: %s" % (e.__class__.__name__, summary_dict["_task_index"], e))
            else:
                logger.error(
                    "[%s] writing trace for task ???: %s" % (e.__class__.__name__, e))
            raise

    def _write_trace(self, summary_dict):

        summary_line = "%d\t%s\t" % (
            summary_dict["_task_index"],
            summary_dict["_status"]
            )

        if 'optimizer' in self.params.vals and self.params.vals["optimizer"] in ["spearmint", "ray_tune"]:
            summary_line += "%f\t" % summary_dict['_score']
            summary_line += "%s\t" % summary_dict["_cmd"]
        else:
            for arg in list(summary_dict["_cmd"].values()):
                summary_line += "%s\t" % str(arg['value'])

        summary_line += "%d\t%s\t%s\t%s" % (
            summary_dict["_return_code"],
            summary_dict["_duration"],
            summary_dict["_start"],
            summary_dict["_end"]
        )

        if "summary" in self.params.vals:
            for summary_val in self.params.vals['summary']:
                summary_line += '\t'
                summary_line += summary_dict[summary_val['id']]

        summary_line = summary_line + "\n"

        trace_path = os.path.join(self.logdir, 'trace.txt')
        trace_file = open(trace_path, 'a')
        trace_file.write(summary_line)
        trace_file.close()

    def wrap_cmd(self, task_index, cmd):
        """Wrap low-level execution of a command with trace and summary handling."""
        try:
            return self._wrap_cmd(task_index, cmd)
        except Exception as e:
            logger.error("[%s] running command for task %d: %s" % (e.__class__.__name__, task_index, e))
            raise

    def _wrap_cmd(self, task_index, cmd):

        logger.debug(" - Opening .out/.err files for task %s" % task_index)
        stdout_path = os.path.join(self.logdir, "%d.out" % task_index)
        stderr_path = os.path.join(self.logdir, "%d.err" % task_index)
        stderr_file = open(stderr_path, 'w')
        stdout_file = open(stdout_path, 'w')

        start = datetime.datetime.now()
        logger.debug(" - Task %s start time is %s", task_index, start)
        ret_code = self._run_cmd(cmd, stderr_file, stdout_file)
        end = datetime.datetime.now()
        logger.debug(" - Task %s end time is %s", task_index, end)
        duration = end - start
        logger.debug(" - Task %s duration is %s", task_index, duration)

        logger.debug(" - Closing .out/.err files for task %s" % task_index)
        stderr_file.close()
        stdout_file.close()

        summary_dict = {
            "_cmd": cmd,
            "_task_index": task_index,
            "_return_code": ret_code,
            "_duration": duration,
            "_start": start,
            "_end": end,
            "_status": "COMPLETE",
        }

        if ret_code != 0:
            logger.warning(" - Task %s had a non-zero exit code, may have failed" % ret_code)
            summary_dict["_status"] = "ERROR"

        if "summary" in self.params.vals:
            for summary_val in self.params.vals['summary']:
                val = self.__pygrep_first(summary_val['regex'], stdout_path)
                if val is None:
                    summary_dict[summary_val['id']] = "ERROR"
                    summary_dict["_status"] = "ERROR"
                else:
                    summary_dict[summary_val['id']] = val

        #
        # Since a default score was set to 'Accuracy' in param.py, this if is not necessary anymore.
        # But, will keep it till a better way to deal with the score was found.
        #
        if "score" in self.params.vals:
            if self.params.vals["score"] in summary_dict:
                score_name = self.params.vals["score"]
                score_val = summary_dict[score_name]
                try:
                    summary_dict["_score"] = float(score_val)
                    logger.debug(" - Score for task %d: %0.4f", summary_dict["_task_index"], summary_dict["_score"])
                    self._write_trace(summary_dict)
                except ValueError:
                    summary_dict["_score"] = 1e99
                    summary_dict["_status"] = "ERROR"
                    self._write_trace(summary_dict)
                    raise ValueError("Could not extract a numerical score from output '%s'" % score_val)
            else:
                summary_dict["_score"] = 1e99
                summary_dict["_status"] = "ERROR"
                self._write_trace(summary_dict)
                raise ValueError("score %s is not in the list of summary variables" % self.params["score"])

        # spearmint/ray_tune score
        if 'optimizer' in self.params.vals.keys() and self.params.vals["optimizer"] in ["spearmint"]:
            logger.debug('{0} optimizor is used'.format(self.params.vals['optimizer']))
            try:
                fcon = open(stderr_path).readlines()
                if self.executor in ['local']:
                    best_value = float([i.split()[2] for i in fcon if i.startswith('Current best:')][-1])
                else:
                    best_value = 1e99
                status = "COMPLETE"
            except Exception as e:
                best_value = 1e99
                status = "ERROR"
                logger.info("Can't get the best_value, the error is: ", e)

            summary_dict["_score"] = best_value
            summary_dict["_status"] = status
            try:
                self._write_trace(summary_dict)
            except Exception as e:
                logger.debug("The trace error is:", e)

            logger.debug('The best value and status are: %f, %s' % (best_value, status))

        elif 'optimizer' in self.params.vals.keys() and self.params.vals["optimizer"] in ["ray_tune"]:
            logger.debug('{0} optimizor is used'.format(self.params.vals['optimizer']))
            try:
                fcon = open(stdout_path).readlines()
                if self.executor in ['local']:
                    best_value = sorted([float(i.strip().split()[-2]) for i in fcon if i.find('| TERMINATED |') > 0])[0]
                else:
                    best_value = 1e99
                status = "COMPLETE"
            except Exception as e:
                best_value = 1e99
                status = "ERROR"
                logger.info("Can't get the best_value, the error is: ", e)

            summary_dict["_score"] = best_value
            summary_dict["_status"] = status
            try:
                self._write_trace(summary_dict)
            except Exception as e:
                logger.debug("The trace error is:", e)

            logger.debug('The best value and status are: %f, %s' % (best_value, status))

        logger.debug('The summary_dict is:')
        logger.debug(summary_dict)
        logger.debug('')
        logger.debug('The self.params.vals are:')
        logger.debug(self.params.vals)
        logger.debug('')

        return summary_dict

    @staticmethod
    def __pygrep_first(regex, text_file):

        for i, line in enumerate(open(text_file)):
            for match in re.finditer(regex, line):
                return match.group(1)

        return None

    @abstractmethod
    def _run_cmd(self, cmd, stderr_file, stdout_file):
        pass
