"""Executors for local (single node) task execution."""

import os
import shutil
import logging
import multiprocessing
from subprocess import call
from subprocess import Popen

from .base import BaseExecutor

logger = logging.getLogger(__name__)


class LocalExecutor(BaseExecutor):
    """LocalExecutor will run tasks on a single local host.

    A multiprocessing pool is used, which will run N tasks,
    where N is (CPUs on the host) / (cpus_per_task in .yml file)
    """

    def __init__(self, cwd, params):
        """Initialize local execution, computing concurrent tasks to run."""
        # set default log level as debug
        # logger.setLevel(logging.DEBUG)
        logger.info(" - Initializing pool for local tasks")
        cpus = multiprocessing.cpu_count()
        logger.info(" - %d CPUs available on this machine" % cpus)
        cpus_per_task = params.vals['cpus_per_task']
        self.conc_tasks = int(cpus // cpus_per_task)

        if self.conc_tasks < 1:
            raise ValueError("Cannot run any tasks - cpus_per_task (%d) is larger than CPUs on this machine."
                             % cpus_per_task)

        if 'optimizer' in params.vals.keys() and params.vals["optimizer"] == "spearmint":
            # max concurrent tasks
            if 'spearmint_max_concurrent_tasks' not in params.vals.keys():
                params.vals["spearmint_max_concurrent_tasks"] = self.conc_tasks
            elif params.vals["spearmint_max_concurrent_tasks"] > self.conc_tasks:
                params.vals["spearmint_max_concurrent_tasks"] = self.conc_tasks

            logger.info(" - Will run %d concurrent local tasks" % params.vals["spearmint_max_concurrent_tasks"])

            # drivers
            if params.vals['nodes'] != 1:
                logger.warning("Local executor can only run on 1 node, reset 'nodes' to '1' and 'driver' to 'local'")
                params.vals['nodes'] = 1

        if 'optimizer' in params.vals.keys() and params.vals["optimizer"] == "ray_tune":
            # drivers
            if params.vals['nodes'] != 1:
                logger.warning("Local executor can only run on 1 node, reset 'nodes' to '1' and 'driver' to 'local'")
                params.vals['nodes'] = 1

        super().__init__(cwd, params)

        self.executor = 'local'

        logger.info(" - Initializing local wrapper for ray_tune tasks")
        self.params.sbatchs = self.__compute_sbatchs()

    def run(self, task_index, cmd):
        """Run provided commands, in parallel up to conc_tasks."""
        logger.debug("LocalExecutor is handling command: %s", cmd)
        return self.wrap_cmd(task_index, cmd)

    def _run_cmd(self, cmd, stderr_file, stdout_file):
        """Local execution of a command under a simple shell call."""
        if 'optimizer' in self.params.vals.keys() and self.params.vals["optimizer"] == "spearmint":

            # check if input files are all here
            logger.info(" - checking required files for this task")

            required_files = [self.params.vals['spearmint_config_file'], self.params.vals['spearmint_function_file']]
            for rf in required_files:
                if not os.path.exists(rf):
                    logger.error("The function file '{0}' is missing.".format(rf))
                    raise FileNotFoundError

            # run the command
            logger.debug("Running: %s\nCWD: %s" % (cmd, os.path.abspath('.')))
            p = Popen(cmd.split(), stderr=stderr_file, stdout=stdout_file)
            logger.debug("The PID of the cmd is: ")
            logger.debug(p.pid)
            p.wait()
            ret = p.returncode
            logger.debug("The return code of the cmd is: ")
            logger.debug(ret)
            p.kill()
        elif 'optimizer' in self.params.vals.keys() and self.params.vals["optimizer"] == "ray_tune":
            sbatch_file = 'param_runner_{0}.sh'.format(self.params.vals['optimizer'])

            # check if input files are all here
            logger.info(" - checking required files for task, %s", sbatch_file)

            required_files = [self.params.vals['rt_function_file']]
            for rf in required_files:
                if not os.path.exists(rf):
                    logger.error("The function file '{0}' is missing.".format(rf))
                    raise FileNotFoundError

            # writing out batch file and generate the command
            logger.info(" - writing out batch file for task, %s", sbatch_file)
            open(sbatch_file, 'w').write(self.params.sbatchs[0])
            cmd = "{0} {1}".format(shutil.which('bash'), sbatch_file)

            # run the command
            logger.debug("Running: %s\nCWD: %s" % (cmd, os.path.abspath('.')))
            p = Popen(cmd.split(), stderr=stderr_file, stdout=stdout_file)
            logger.debug("The PID of the cmd is: ")
            logger.debug(p.pid)
            p.wait()
            ret = p.returncode
            logger.debug("The return code of the cmd is: ")
            logger.debug(ret)
            p.kill()
        else:
            logger.debug("Running: %s" % cmd['__command']['value'])
            ret = call(cmd['__command']['value'].split(), shell=False, stderr=stderr_file, stdout=stdout_file)
        return ret

    def __compute_sbatchs(self):
        """Compute the sbatch scripts to be submitted to the cluster.

        Returns:
            a list of sbatchs

        """
        logger.info("Generating SBATCH ...")
        logger.debug(self.params)

        if 'optimizer' in self.params.vals.keys() and self.params.vals["optimizer"] == "spearmint":
            arg_combns = self.__compute_sbatchs_optimizer(optimizer='spearmint')
        elif 'optimizer' in self.params.vals.keys() and self.params.vals["optimizer"] == "ray_tune":
            arg_combns = self.__compute_sbatchs_optimizer(optimizer='ray_tune')
        else:
            arg_combns = ['']
            logger.warning("compute batchs have not been implemented yet")

        return arg_combns

    def __compute_sbatchs_optimizer(self, optimizer="spearmint"):
        arg_combns = "#!/bin/bash\n"

        if optimizer == "spearmint":
            arg_combns += "{0}\n".format(self.params.commands[0])

        if optimizer == "ray_tune":
            s = """
Input="{0}"

# Localdir will be used by $Input
export Localdir="/tmp/param_runner_{1}/$$"
mkdir -p $Localdir
chmod 777 $(dirname $Localdir)
Outdir="$(pwd)/$(echo $Input | awk -F. '{{print $1}}')"

# non empty proxy settings will cause connection issue!
unset http_proxy
unset https_proxy

# get hostname and num of cpus
redis_port=6379
ip_prefix=$(hostname --ip-address)
NUM_CPUs=$(grep "^processor" /proc/cpuinfo | wc -l)

echo "Head node: $(hostname)"
export RAY_HEAD_IP="$ip_prefix:$redis_port"
export RAY_REDIS_PASSWD=$(uuidgen)

# get the num of GPUs, and set the --num-gpus flag for ray
if [[ -z $CUDA_VISIBLE_DEVICES ]]; then
    NUM_GPUs=0
    RAY_GPUs=""
    CUDA_VISIBLE_DEVICES=''
elif [ $CUDA_VISIBLE_DEVICES = 'NoDevFiles' ] || [ $CUDA_VISIBLE_DEVICES = '' ]; then
    NUM_GPUs=0
    RAY_GPUs=""
    CUDA_VISIBLE_DEVICES=''
else
    NUM_GPUs=$(echo $CUDA_VISIBLE_DEVICES  | sed "s/,/\\n/g" | wc -l )
    RAY_GPUs="--num-gpus=$NUM_GPUs"
fi

# NUM_GPUs will be used by $Input
export NUM_GPUs

# Starting the head
ray start --head --redis-port=$redis_port --redis-password=$RAY_REDIS_PASSWD &
sleep 5

# start the calculation
python3 $Input

# stop ray
ray stop

# collect the results
mkdir -p $Outdir
cp -r $Localdir/* $Outdir/
rm -fr $Localdir
            """.format(self.params.vals['rt_function_file'], optimizer)

            arg_combns += s
        return [arg_combns]
