"""Executors  / utility code for task execution on a SLURM cluster."""

import logging
import os
import re
import sys
from subprocess import call

from .base import BaseExecutor

logger = logging.getLogger(__name__)


class SrunExecutor(BaseExecutor):
    """SrunExecutor will run tasks within an existing SLURM allocation.

    Tasks are dispatched into a SLURM allocation using srun. The number of
    concurrent tasks is controlled via a multiprocessing pool, and computed
    from SLURM environment variables for CPU / node counts, and the specified
    cpus_per_task value in the YAML parameter file.
    """

    SRUN_DEFAULT_ARGS = "--exclusive -N1 -n1 --distribution=cyclic"
    """Default arguments to the srun command. CPU / GRES requirements will be
    appended.
    Each task run by the parameter runner is assumed to run on a single node (-N1),
    is a single non-MPI task (-n1), and allocation is set to cyclic to evenly
    allocated tasks across nodes rather than filling first node first.
    """

    srun_args = None

    def __init__(self, cwd, params):
        """Initialize srun execution, computing concurrent tasks to run."""
        if 'SLURM_JOB_ID' not in os.environ:
            raise EnvironmentError(
                'The srun executor can only be run inside a SLURM allocation.')

        logger.info(" - Initializing pool for srun tasks")

        logger.info(
            " - Running in SLURM JOB %s" % os.environ.get('SLURM_JOB_ID'))
        cpus_per_node = int(
            re.match(r"\d+", os.environ.get('SLURM_JOB_CPUS_PER_NODE')).group(0))
        logger.info(" - %d CPUs per node" % cpus_per_node)
        nodes = int(os.environ.get('SLURM_NNODES'))
        logger.info(" - %d nodes" % nodes)
        cpus_per_task = params.vals['cpus_per_task']
        self.conc_tasks = (cpus_per_node / cpus_per_task) * nodes
        logger.info(" - Will run %d Concurrent tasks" % self.conc_tasks)

        # Initialize SRUN arguments
        self.srun_args = self.SRUN_DEFAULT_ARGS
        self.srun_args = self.srun_args + " --cpus-per-task=%d" % params.vals['cpus_per_task']

        gres_spec = ""
        if 'gpus_per_task' in params.vals:
            gres_spec = gres_spec + "gpu:%d" % params.vals['gpus_per_task']
            logger.info(" - Requesting %d GPUs per task (%s)",
                        params.vals['gpus_per_task'], gres_spec)

        if 'gres_per_task' in params.vals:
            if gres_spec != "":
                gres_spec = gres_spec + ","
            gres_spec = gres_spec + "gpu:%d" % params.vals['gres_per_task']
            logger.info(" - Requesting %s gres per task", gres_spec)

        if gres_spec != "":
            self.srun_args = self.srun_args + " --gres=" + gres_spec

        if 'timelimit_per_task' in params.vals:
            self.srun_args = self.srun_args + " -t " + params.vals['timelimit_per_task']
            logger.info(" - Time limit per task is %s")

        logger.debug(" - srun arguments: %s", self.srun_args)

        super().__init__(cwd, params)

        self.executor = 'srun'

    def run(self, task_index, cmd):
        """Run provided commands, in parallel up to conc_tasks."""
        logger.debug("SrunExecutor is handling command: %s", cmd)
        return self.wrap_cmd(task_index, cmd)

    def _run_cmd(self, cmd, stderr_file, stdout_file):
        """SLURM execution of a command using srun."""
        srun_cmd = "srun %s %s" % (self.srun_args, cmd['__command']['value'])
        logging.debug(" - srun command for task is: %s", srun_cmd)

        ret = call(srun_cmd.split(), shell=False, stderr=stderr_file,
                   stdout=stdout_file, env=os.environ)
        return ret


class SlurmWrapperExecutor(BaseExecutor):
    """SlurmWrapperExecutor will provide Slurm scripts for running the job."""

    def __init__(self, cwd, params):
        """Initialize srun execution, computing concurrent tasks to run."""
        super().__init__(cwd, params)

        if 'optimizer' in params.vals.keys() and params.vals["optimizer"] == "spearmint":
            # drivers
            if params.vals['nodes'] != 1:
                logger.warning("Spearmint can only run on 1 node, reset 'nodes' to '1' and 'driver' to 'local'")
                params.vals['nodes'] = 1

        self.executor = 'slurm'
        self.sbatch_file = 'param_runner_{0}.sbatch'.format(self.params.vals['optimizer'])

        logger.info(" - Initializing SLURM wrapper for tasks")
        self.params.sbatchs = self.__compute_sbatchs()

    def run(self, task_index, cmd):
        """Run provided commands, in parallel up to conc_tasks."""
        logger.debug("SrunExecutor is handling command: %s", cmd)
        return self.wrap_cmd(task_index, cmd)

    def _run_cmd(self, cmd, stderr_file, stdout_file):
        logging.info(" - checking for the required files for task, %s", self.sbatch_file)

        # check for the input files
        if self.params.vals['optimizer'] in ['spearmint']:
            # test for config and function file
            config_file = self.params.vals['spearmint_config_file']
            func_file = self.params.vals['spearmint_function_file']
            if not os.path.exists(config_file):
                logger.error("Could not find spearmint_config_file: %s" % config_file)
                raise FileNotFoundError
            if not os.path.exists(func_file):
                logger.error("Could not find spearmint_function_file: %s" % func_file)
                raise FileNotFoundError
        elif self.params.vals['optimizer'] in ['ray_tune']:
            # test for function file
            func_file = self.params.vals['rt_function_file']
            if not os.path.exists(func_file):
                logger.error("Could not find rt_function_file: %s" % func_file)
                raise FileNotFoundError

        # write the sbatch file
        logging.info(" - writing out sbatch file for task, %s", self.sbatch_file)
        try:
            open(self.sbatch_file, 'w').write(self.params.sbatchs[0])
            return 0
        except Exception as e:
            logger.debug("Failed to write sbatch file: %s" % self.sbatch_file)
            logger.debug(e)

    def __compute_sbatchs(self):
        """Compute the sbatch scripts to be submitted to the cluster.

        Returns:
            a list of sbatchs

        """
        logger.info("Generating SBATCH ...")
        logger.debug(self.params)

        if 'optimizer' in self.params.vals.keys() and self.params.vals["optimizer"] == "spearmint":
            arg_combns = self.__compute_sbatchs_optimizer(optimizer='spearmint')
        elif 'optimizer' in self.params.vals.keys() and self.params.vals["optimizer"] == "ray_tune":
            arg_combns = self.__compute_sbatchs_optimizer(optimizer='ray_tune')
        else:
            arg_combns = ['']
            logger.warning("compute sbatchs have not been implemented yet")

        return arg_combns

    def __compute_sbatchs_optimizer(self, optimizer="spearmint"):
        if 'gpus_per_task' not in self.params.vals.keys():
            num_gpus = 0
        else:
            num_gpus = self.params.vals['gpus_per_task']

        arg_combns = "#!/bin/bash\n" \
                     "#SBATCH --job-name=param_runner_{0}\n" \
                     "#SBATCH --partition={1}\n" \
                     "#SBATCH --nodes={2}\n" \
                     "#SBATCH --ntasks-per-node={3}\n" \
                     "#SBATCH --gres=gpu:{4}\n" \
                     "#SBATCH --time={5}\n" \
                     "#SBATCH --output=param_runner.%j.out\n" \
                     "#\n" \
                     "# This script was generated by param_runner\n" \
                     "#\n" \
                     "# Command to run\n".format(optimizer,
                                                 self.params.vals['partition'],
                                                 self.params.vals['nodes'],
                                                 self.params.vals['cpus_per_task'],
                                                 num_gpus,
                                                 self.params.vals['time_limit'])

        if optimizer == "spearmint":
            arg_combns += "{0}\n".format(self.params.commands[0])

        if optimizer == "ray_tune":
            s = """
Input="{0}"

# Localdir will be used by $Input
export Localdir="/tmp/param_runner_{1}/$$"
mkdir -p $Localdir
chmod 777 $(dirname $Localdir)
Outdir="$(pwd)/$(echo $Input | awk -F. '{{print $1}}')"

## load module
#module load python/3.7.x-anaconda
## start the virtual environment
#source activate ray

export PATH={2}:$PATH

# non empty proxy settings will cause connection issue!
unset http_proxy
unset https_proxy

# Must be one less that the total number of nodes
worker_num=$(echo "$SLURM_JOB_NUM_NODES - 1" | bc -l)

# Getting the node names
nodes=$(scontrol show hostnames $SLURM_JOB_NODELIST)
nodes_array=( $nodes )

# get 1st node
node1=${{nodes_array[0]}}
echo "Head node: $node1"

# get head node ip and port
redis_port=6379
ip_prefix=$(srun --nodes=1 --ntasks=1 -w $node1 hostname --ip-address)

export RAY_HEAD_IP="$ip_prefix:$redis_port"
export RAY_REDIS_PASSWD=$(uuidgen)

# get the num of GPUs, and set the --num-gpus flag for ray
if [[ -z $CUDA_VISIBLE_DEVICES ]]; then
    NUM_GPUs=0
    RAY_GPUs=""
    CUDA_VISIBLE_DEVICES=''
elif [ $CUDA_VISIBLE_DEVICES = 'NoDevFiles' ] || [ $CUDA_VISIBLE_DEVICES = '' ]; then
    NUM_GPUs=0
    RAY_GPUs=""
    CUDA_VISIBLE_DEVICES=''
else
    NUM_GPUs=$(echo $CUDA_VISIBLE_DEVICES  | sed "s/,/\\n/g" | wc -l )
    RAY_GPUs="--num-gpus=$NUM_GPUs"
fi

# NUM_GPUs will be used by $Input
export NUM_GPUs

# Starting the head
srun --nodes=1 --ntasks=1 -w $node1 ray start --block --head \
     --redis-port=$redis_port --redis-password=$RAY_REDIS_PASSWD &
sleep 5
# Make sure the head successfully starts before any worker does, otherwise
# the worker will not be able to connect to redis. In case of longer delay,
# adjust the sleeptime above to ensure proper order.

for ((  i=1; i<=$worker_num; i++ ))
do
  node2=${{nodes_array[$i]}}
  echo "Starting work nodes: $node2"
  srun --nodes=1 --ntasks=1 -w $node2 ray start --block --address=$RAY_HEAD_IP --redis-password=$RAY_REDIS_PASSWD &
  # Flag --block will keep ray process alive on each compute node.
  sleep 5
done

# start the calculation
python3 $Input

# collect the results
mkdir -p $Outdir
cp -r $Localdir/* $Outdir/
rm -fr $Localdir
            """.format(self.params.vals['rt_function_file'], optimizer, os.path.dirname(sys.executable))

            arg_combns += s
        return [arg_combns]
