"""Base class and utility functions for Executors."""

import logging
from abc import ABCMeta, abstractmethod

logger = logging.getLogger(__name__)


class BaseOptimizer(object):
    """Base class for an optimizer.

    An Optimizer is responsible for exploring a parameter space by generating
    tasks that will be run by an executor, (optionally) guided by scores
    returned from those tasks.

    An *Optimizer subclass must implement a single run method, which should
    carry out the entire optimization flow, using the parameters and executor
    provided on __init__.
    """

    metaclass = ABCMeta

    params = None
    """Representation of optimization parameters from the .yml file"""

    executor = None
    """Subclass of BaseExecutor which can execute tasks on behalf of
    the optimizer."""

    def __init__(self, params, executor):
        """Initialize the optimizer, with a provided parameter set and executor."""
        logger.info("Initializing %s " % self.__class__.__name__)

        self.params = params
        self.executor = executor

    @abstractmethod
    def run(self):
        """Run an optimization.

        Must be implemented to carry out the entire optimization flow,
        using the parameters and executor  provided on __init__.
        """
        pass
