"""Basic optimization algorithm implementations."""

import logging

from .base import BaseOptimizer

logger = logging.getLogger(__name__)
# set default log level as debug
# logger.setLevel(logging.DEBUG)


class GridSearchOptimizer(BaseOptimizer):
    """Exhaustive grid search across parameter ranges."""

    task_index = 0
    best_task = {
        'task_index': -1,
        'score': 1e99
    }

    def run(self):
        """Run a grid search, passing all tasks to the executor immediately."""
        tasks_total = len(self.params.commands)
        logger.info(" - Exhaustive grid search will run %d tasks" % tasks_total)

        pool = self.executor.get_pool()

        for task in self.params.commands:
            pool.apply_async(self.run_job, [self.task_index, task], callback=self.score_check)
            self.task_index = self.task_index + 1

        pool.close()
        pool.join()

        logger.info(" - Optimization completed after %d tasks run" % tasks_total)

        logger.info(" - Best score: %f\ttask %d", self.best_task['score'],
                    self.best_task['task_index'])

    def run_job(self, task_index, cmd):
        """Run a single task within the grid search optimization."""
        try:

            logger.info(" - Starting task %d", task_index)
            summary = self.executor.run(task_index, cmd)
            logger.info(" - Finished task %d score: %f", task_index, summary["_score"])

            return {
                'task_index': task_index,
                'score': summary["_score"]
            }

        except Exception as e:
            logger.error("Could not run task %d: %s", task_index, e)
            return None

    def score_check(self, task_result):
        """Check if the score returned by a task is a new minimum."""
        if task_result and task_result['score'] < self.best_task['score']:
            self.best_task = task_result
            logger.info(" - New best score: %f\ttask %d", task_result['score'],
                        task_result['task_index'])
