"""Implementation of the Spearmint optimization algorithm."""

import logging

from .base import BaseOptimizer


logger = logging.getLogger(__name__)


class TuneOptimizer(BaseOptimizer):
    """Implementation of the Ray Tune optimization."""

    task_index = 0
    best_task = {
        'task_index': None,
        'job_id': None,
        'score': 1e99
    }

    def run(self):
        """Pass all tasks to the executor immediately."""
        task_result = None
        tasks_total = len(self.params.commands)
        logger.info(" - Will run %d tasks" % tasks_total)
        if self.executor.executor == 'local':
            for task in self.params.commands:
                task_result = self.run_job(self.task_index, task)
                self.score_check(task_result)

                self.task_index = self.task_index + 1

            # report the score if using local executor
            logger.info(" - Best score: %f\ttask %d", self.best_task['score'],
                        self.best_task['task_index'])
        else:
            for task in self.params.commands:
                task_result = self.run_job(self.task_index, task)

                self.task_index = self.task_index + 1

        if task_result:
            logger.info(" - Optimization completed after %d tasks run" % tasks_total)
            return True
        else:
            return False

    def run_job(self, task_index, cmd):
        """Run a single task."""
        try:

            logger.info(" - Starting task %d", task_index)
            summary = self.executor.run(task_index, cmd)

            if self.executor.executor in ['local']:
                ret_dict = {
                    'task_index': task_index,
                    'score': summary["_score"]
                }
                logger.info(" - Finished task %d score: %f", task_index, summary["_score"])

            else:
                ret_dict = {'task_index': task_index}
                logger.info(" - Finished task %d.", task_index)

            return ret_dict

        except Exception as e:
            logger.error("Could not run task %d: %s", task_index, e)
            return None

    def score_check(self, task_result):
        """Check if the score returned by a task is a new minimum."""
        if task_result and task_result['score'] < self.best_task['score']:
            self.best_task = task_result
            logger.info(" - New best score: %f\ttask %d", task_result['score'],
                        task_result['task_index'])
