"""The param module hold classes related to a parameter_runner configuration file."""
import logging
import os
import subprocess
import re
import shutil
from collections import OrderedDict
from copy import deepcopy
from decimal import Decimal

import anyconfig
from yamllint import linter
from yamllint.cli import Format
from yamllint.config import YamlLintConfig

logger = logging.getLogger(__name__)
# set default log level as debug
# logger.setLevel(logging.DEBUG)


# PATH TO MODULECMD
# TODO: make configurable nicely
# MODULECMD = "/cm/local/apps/environment-modules/current/bin/modulecmd"
MODULECMD = shutil.which('modulecmd')


class ParamFileException(Exception):
    """Exception for any errors related to the parameter file."""

    pass


class ParamFile(object):
    """Process and hold the parameter file that defines the experiment to run.

    A parameter file is a YAML file, which defines:

      * An executable we want to run with different parameters
      * Generic options for the cluster job we will create
      * The paremeter ranges to explore

    This class validates parameter files and ingests them. The validation is
    performed using a jsonschema format custom schema, plus the yamllint tool
    with customized rules to silence minor warnings.

    Attributes:
        param_file (str): path to the input parameter yaml file
        vals (dict): Dictionary of configuration data extracted reflecting yaml
        combination_count (int): Number of parameter combinations
        commands: Array of commands to be run, computed from param ranges

    """

    param_file = ''
    vals = {}
    params = OrderedDict()
    combination_count = None
    commands = []
    sbatchs = []                # sbatch script for slurm
    loaded = False

    mypath = os.path.dirname(os.path.abspath(__file__))

    # Spearmint is available in the searching path, use it. Otherwise, use the one in this package
    if shutil.which('spearmint'):
        spearmint = shutil.which('spearmint')
    else:
        spearmint = os.path.join(os.path.dirname(mypath), "param_runner", "lib", "spearmint", "bin", "spearmint")

    schema_file = os.path.join(mypath, 'schema', 'param_runner.jsonschema.yml')
    lint_config = os.path.join(mypath, 'schema', 'yamllint.yml')

    # anyconfig.*load*() do not provide a way to provide default configuration values.
    # https://pythonhosted.org/anyconfig/usage.html
    defaults_in_schema = {
        'nodes': 1,
        'spearmint_method': 'GPEIOptChooser',
        'spearmint_port': 15272,
        'spearmint_host': '127.0.0.1',
    }

    def __init__(self, param_file):
        """Initialize, checking for existence of parameter file only."""
        if not os.path.exists(param_file):
            raise IOError("Parameter file %s not found." % param_file)

        self.param_file = param_file

    def validate(self):
        """Validate our parameter file against the jsonschema definition.

        This uses anyconfig, which verifies structure, but can give obscure
        errors on syntax - so preferably use it after an initial screen for
        syntax with yamllint via lint method.

        Returns:
            True if metadata valid. Raises an exception on a validation error.

        """
        logger.info(
            "Validating Parameter File %s against schema" % self.param_file)

        params = anyconfig.load(self.param_file)
        schema = deepcopy(self.defaults_in_schema)
        schema.update(anyconfig.load(self.schema_file))

        (rc, err) = anyconfig.validate(params, schema)

        if err:
            message = "Parameter file invalid: " + err.splitlines()[0]
            raise ParamFileException(message)

        logger.info("Parameter file is valid")

        return self.checkModules(params)

    def checkModules(self, params):
        """Check modules specified for the workflow exist.

        Use the environment modules system modulecmd to check any modules
        requested for this run actually exist on the system
        """
        if not MODULECMD:
            modules_ok = True               # set validation as True as 'modules' is not required in some cases
            logger.warning("Could not find 'modulecmd'. No modules will be loaded.")
            return modules_ok

        modules_ok = True
        if 'modules' in list(params.keys()):

            for mod_name in params['modules']:
                cmd = [MODULECMD, 'sh', 'show', mod_name]
                output = subprocess.check_output(cmd, stderr=subprocess.STDOUT)

                if "Unable to locate" in str(output):
                    logger.error('Module %s was not found.' % mod_name)
                    modules_ok = False
                else:
                    logger.info('Module %s will be loaded.' % mod_name)

        return modules_ok

    def lint(self):
        """Lint our parameter file with yamllint.

        anyconfig with a jsonschema gives unfriendly errors for standard syntax
        issues. We use yamllint with a customized ruleset defined in
        'schema/yamllint.yml' to give friendlier output for common issues.

        Returns:
            True if metadata valid. Raises an exception on a critical error.

        """
        logger.info('Linting Parameter File %s' % self.param_file)

        conf = YamlLintConfig(file=self.lint_config)
        err = 0

        with open(self.param_file) as f:

            for problem in linter.run(f, conf):

                logger.warning(Format.standard(problem, ''))

                if err == 0 and problem.level == 'error':
                    err = 1

        if err:
            message = "Parameter file (%s) is invalid. Please fix the errors shown above." % self.metadata_file
            raise ParamFileException(message)

    def load(self):
        """Read in metadata from the astrocyte_pkg.yml to populate vals dict."""
        if self.loaded:
            raise ParamFileException("Parameter information was already loaded.")

        self.validate()
        metadata = deepcopy(self.defaults_in_schema)
        metadata.update(anyconfig.load(self.param_file))
        self.vals = metadata
        if "summary" in self.vals:
            self.__check_summary_regexes()
            self.__check_score()
        if "optimizer" in self.vals and self.vals["optimizer"] in ["spearmint"]:
            self.__compute_param_ranges()
        elif "optimizer" in self.vals and self.vals["optimizer"] in ["ray_tune"]:
            pass
        else:
            self.__compute_param_ranges()

        self.commands = self.__compute_commands()
        if self.__clean_vals():
            self.loaded = True
        else:
            self.loaded = False
            raise ParamFileException("Failed to load parameter information.")

    def __clean_vals(self):
        # clean the spearmint defaults
        if "optimizer" in self.vals and self.vals["optimizer"] in ["spearmint"]:
            pass
        else:
            for k in self.defaults_in_schema.keys():
                if k.startswith('spearmint'):
                    self.vals.pop(k)

        if self.vals['nodes'] > 1:
            if 'partition' in self.vals and 'time_limit' in self.vals:
                return True
            else:
                logger.error('To run on multiple nodes "partition" and "time_limit" are required.')
                return False
        else:
            return True

    def __check_summary_regexes(self):
        for summary in self.vals['summary']:

            try:
                compiled_re = re.compile(summary['regex'])

                if compiled_re.groups != 1:
                    raise TypeError

                logger.info('- Summary %s using regex %s' % (summary['id'], summary['regex']))

            except re.error:
                raise re

    def __check_score(self):

        if "score" in self.vals:
            for summary in self.vals['summary']:
                if summary['id'] == self.vals['score']:
                    logger.info(" - Using '%s' as score to minimize" % summary['id'])
                    return
            raise ValueError("Specified score '%s' is not a defined summary" % self.vals['score'])
        #
        # Set the default score to Accuracy
        # TODO: need a better way to obtain the input score!
        #
        else:
            logger.warning(" - No score summary was provided. Using 'Accuracy' as default.")
            self.vals['score'] = 'Accuracy'

        # logger.warning(" - No score summary was provided")

    def __compute_param_ranges(self):
        """Compute the parameter ranges as lists, from their definitions."""
        logger.info("Computing Parameter Ranges...")

        if 'parameters' in self.vals.keys():
            self.combination_count = 1

            for param in self.vals['parameters']:

                if param['type'] == 'int_range':
                    ir = IntRange(param)
                    param['range'] = ir.to_list()

                if param['type'] == 'real_range':
                    rr = RealRange(param)
                    param['range'] = rr.to_list()

                if param['type'] == 'choice':
                    cr = ChoiceRange(param)
                    param['range'] = cr.to_list()

                self.combination_count *= len(param['range'])

                logger.info(" - Parameter '%s' will explore values: %s " %
                            (param['id'], param['range']))

            logger.info(
                " - Total combinations to explore: %d" % self.combination_count)

    def __compute_commands(self):
        """Compute the commands to be run across parameter ranges.

        Returns:
            cmd_arr(list): List of string commands to be executed over param space

        """
        if self.commands:
            logger.debug(self.commands)
            raise ParamFileException("Commands have already been computed")

        logger.info("Computing Commands...")
        if 'optimizer' in self.vals.keys() and self.vals["optimizer"] == "spearmint":
            arg_combns = self.__compute_commands_optimizer(optimizer='spearmint')
        elif 'optimizer' in self.vals.keys() and self.vals["optimizer"] == "ray_tune":
            arg_combns = self.__compute_commands_optimizer(optimizer='ray_tune')
        else:
            arg_combns = self.__recurse_arg_combns(0, OrderedDict(), [])
            logger.info(" - Computed %d argument combinations" % len(arg_combns))

            assert len(arg_combns) == self.combination_count

            for arg_combn in arg_combns:
                arg_combn['__command'] = {'value': self.cmd_from_args(arg_combn)}

            logger.info(" - Generated %d commands" % len(arg_combns))

        return arg_combns

    def __compute_commands_optimizer(self, optimizer='spearmint'):
        cmd_srt = ''
        if optimizer == 'spearmint':
            sp_optional_cmd_mapping = [
                ('spearmint_max_concurrent_tasks', 'max-concurrent'),
                ('spearmint_max_tasks', 'max-finished-jobs'),
                ('spearmint_method_args', 'method-args'),
                ('spearmint_grid_size', 'grid-size'),
                ('spearmint_grid_seed', 'grid-seed'),
                ('spearmint_polling_time', 'polling-time'),
            ]

            # base command string
            cmd_srt = "%s ./%s --driver=local --method=%s " % \
                      (self.spearmint, self.vals['spearmint_config_file'], self.vals['spearmint_method'])
            # options
            for cmd in sp_optional_cmd_mapping:
                if cmd[0] in self.vals.keys():
                    cmd_srt += "--%s=%s " % (cmd[1], self.vals[cmd[0]])

            # web status interface
            cmd_srt += "-w --port=%d --host=%s" % (self.vals['spearmint_port'], self.vals['spearmint_host'])
        elif optimizer == 'ray_tune':
            # base command string
            cmd_srt = "python %s" % (self.vals['rt_function_file'])

        cmds = [cmd_srt]
        return cmds

    def __recurse_arg_combns(self,
                             param_idx=0,
                             current_combination=OrderedDict(),
                             all_combinations=[]):
        """Recursively find a combinations of arguments for a execution.

        Args:
            param_idx(int): Index of the current parameter explored
                at this recursion level.
            current_comination(dict): Current combination of args being built
                down through the recursive levels.
            all_combinations(list): List of argument combinations dicts computed so
                far.

        Returns:
            all_combinations(list): List of argument combinations dicts computed so
                far.

        """
        params = self.vals['parameters']

        # Tail of recursion - we have a combination so add it to our list
        if param_idx == len(params):
            all_combinations.append(current_combination)
            return all_combinations

        current_param = params[param_idx]
        current_id = current_param['id']
        current_flag = current_param.get('flag', None)
        current_sub = current_param.get('substitution', None)
        num_values = len(current_param['range'])

        for val_idx in range(num_values):
            current_value = current_param['range'][val_idx]
            current_combination[current_id] = {
                'flag': current_flag,
                'substitution': current_sub,
                'value': current_value
            }
            all_combinations = self.__recurse_arg_combns(
                param_idx + 1, deepcopy(current_combination), all_combinations)

        return all_combinations

    def arg_template(self):
        """Generate a dictionary with information about each parameter being explored.

        The 'value' of each parameter is initialized to 'None'. This dictionary is for
        convenience to iterate over in optimizer/executor classes to set values for each
        task execution.

        The dictionary is indexed by the 'id' of the parameter, and will contain:
          - type
          - flag
          - substitution (or None)
          - value (None)
        keys, set where appropriate from the parameter configuration.
        """
        args = {}

        for param in self.vals['parameters']:
            id = param['id']
            args[id] = {}
            args[id]['id'] = id
            args[id]['type'] = param.get('type')
            args[id]['flag'] = param.get('flag', None)
            args[id]['substitution'] = param.get('substitution', None)
            args[id]['value'] = None

        return args

    def cmd_from_args(self, arg_combn):
        """Convert a dict of arguments into a full command string.

        Args:
            arg_combn(dict): Dictionary of arguments with flag and value

        Returns:
            cmd_str(str): Executable command string

        """
        cmd_str = self.vals['command'] + ' '

        for arg in list(arg_combn.values()):
            # In substitution mode we replace a placeholder in the command string
            if arg['substitution'] is not None:
                cmd_str = cmd_str.replace(arg['substitution'], arg['value'])

            # Otherwise we append args, possibly prefixed with specified flags
            else:

                if arg['flag'] and arg['value'] is not None:
                    cmd_str += arg['flag'] + ' '

                if arg['value'] is not None:
                    cmd_str += str(arg['value']) + ' '

        return cmd_str.rstrip(' ')


class IntRange(object):
    """Compute integer parameters from an int_range definition.

    Attributes:
        param (dict): Range definition dictionary

    """

    param = None

    def __init__(self, param):
        """Initialize using the range information passed through from the config YAML file."""
        self.param = param

    def to_list(self):
        """Return a list of all integer values to explore, assuming grid search."""
        min_val = int(self.param['min'])
        max_val = int(self.param['max'])

        step = int(self.param.get('step', 1))
        scale = int(self.param.get('scale', 1))

        # Special case - if only scale specified then assume no step
        if 'scale' in self.param and 'step' not in self.param:
            step = 0

        current = min_val
        vals = []

        if self.param.get('optional', False):
            vals.append(None)

        while current <= max_val:
            vals.append(str(current))
            current = (current * scale) + step

        return vals


class RealRange(IntRange):
    """Compute real-value parameters from an real_range definition.

    Attributes:
        param (dict): Range definition dictionary

    """

    def to_list(self):
        """Return a list of all real values to explore, assuming grid search."""
        min_val = Decimal(str(self.param['min']))
        max_val = Decimal(str(self.param['max']))
        step = Decimal(str(self.param.get('step', '1.0')))
        scale = Decimal(str(self.param.get('scale', '1.0')))

        # Special case - if only scale specified then assume no step
        if 'scale' in self.param and 'step' not in self.param:
            step = Decimal('0.0')

        current = min_val
        vals = []

        if self.param.get('optional', False):
            vals.append(None)

        while current <= max_val:
            vals.append(str(current))
            current = (current * scale) + step

        return vals


class ChoiceRange(object):
    """Compute choice parameters from an choice definition.

    Attributes:
        param (dict): Range definition dictionary

    """

    param = None

    def __init__(self, param):
        """Initialize using the range information passed through from the config YAML file."""
        self.param = param

    def to_list(self):
        """Return a list of all string values to explore."""
        vals = []

        if self.param.get('optional', False):
            vals.append(None)

        vals = vals + (self.param['values'])

        return vals
