import os

import pytest

from param_runner.param import ParamFile, ParamFileException, IntRange, RealRange, \
    ChoiceRange

class TestParams:
    def test_validate(self):
        md = ParamFile(os.path.dirname(
            os.path.abspath(__file__)) + '/test_data/test_params.yaml')
        assert md.validate()

    def test_validate_exception(self):
        md = ParamFile(os.path.dirname(
            os.path.abspath(__file__)) + '/test_data/test_params_invalid.yaml')
        with pytest.raises(ParamFileException):
            md.validate()

    def test_validate_no_summary(self):
        md = ParamFile(os.path.dirname(
            os.path.abspath(__file__)) + '/test_data/test_params_no_summary.yaml')
        assert md.validate()

    def test_validate_spearmint(self):
        md = ParamFile(os.path.dirname(
            os.path.abspath(__file__)) + '/test_data/test_params_spearmint.yaml')
        assert md.validate()

    def test_validate_ray_tune(self):
        md = ParamFile(os.path.dirname(
            os.path.abspath(__file__)) + '/test_data/test_params_tune.yaml')
        assert md.validate()

    def test_load(self):
        md = ParamFile(os.path.dirname(
            os.path.abspath(__file__)) + '/test_data/test_params.yaml')
        md.load()
        assert md.vals
        # IMPLICIT md.compute_param_ranges()
        #  min: 0
        #  max: 32
        #  step: 2
        assert md.vals['parameters'][0]['range'] == ['0', '2', '4', '6', '8', '10', '12', '14', '16', '18', '20', '22', '24', '26', '28', '30', '32']
        # optional: true
        # min: 0.1
        # max: 100
        # scale: 10
        assert md.vals['parameters'][1]['range'] == [None, '0.1', '1.0', '10.0',
                                                     '100.0']
        # values:
        #   - step
        #   - sigmoid
        assert md.vals['parameters'][2]['range'] == ['step', 'sigmoid']

    def test_load_spearmint(self):
        md = ParamFile(os.path.dirname(
            os.path.abspath(__file__)) + '/test_data/test_params_spearmint.yaml')
        md.load()
        assert md.vals
        assert md.vals['optimizer'] == 'spearmint'

        for k in ['spearmint_config_file', 'spearmint_function_file', 'spearmint_method',
                  'spearmint_port', 'spearmint_host']:
            assert k in md.vals.keys()

        assert len(md.commands) == 1
        assert md.commands[0].split()[0] == os.path.join(
            os.path.dirname(os.path.dirname(os.path.abspath(__file__))), "lib", "spearmint", "bin", "spearmint")

    def test_load_ray_tune(self):
        md = ParamFile(os.path.dirname(
            os.path.abspath(__file__)) + '/test_data/test_params_tune.yaml')
        md.load()
        assert md.vals
        assert md.vals['optimizer'] == 'ray_tune'

        assert md.vals['gpus_per_task'] == 1

        assert 'rt_function_file' in md.vals.keys()

        assert len(md.commands) == 1
        assert md.commands[0].split()[1] == md.vals['rt_function_file']

    def test_int_range_single(self):
        ir = IntRange({'id': 'test', 'min': 0, 'max': 5})
        assert ir.to_list() == ['0', '1', '2', '3', '4', '5']

    def test_int_range_step(self):
        ir = IntRange({'id': 'test', 'min': 0, 'max': 10, 'step': 2})
        assert ir.to_list() == ['0', '2', '4', '6', '8', '10']

    def test_int_range_scale(self):
        ir = IntRange({'id': 'test', 'min': 1, 'max': 10000, 'scale': 10})
        assert ir.to_list() == ['1', '10', '100', '1000', '10000']

    def test_int_range_step_scale(self):
        ir = IntRange(
            {'id': 'test', 'min': 0, 'max': 10000, 'scale': 10, 'step': 9})
        assert ir.to_list() == ['0', '9', '99', '999', '9999']

    def test_int_range_optional(self):
        ir = IntRange({'id': 'test', 'min': 0, 'max': 5, 'optional': True})
        assert ir.to_list() == [None, '0', '1', '2', '3', '4', '5']

    def test_real_range_single(self):
        rr = RealRange({'id': 'test', 'min': '0.1', 'max': '5.1'})
        assert rr.to_list() == ['0.1', '1.10', '2.100', '3.1000', '4.10000', '5.100000']

    def test_real_range_step(self):
        rr = RealRange({'id': 'test', 'min': '0.1', 'max': '4.9', 'step': '1.2'})
        assert rr.to_list() == ['0.1', '1.30', '2.500', '3.7000', '4.90000']

    def test_real_range_scale(self):
        rr = RealRange({'id': 'test', 'min': '0.001', 'max': '10.0', 'scale': '10'})
        assert rr.to_list() == ['0.001', '0.010', '0.100', '1.000', '10.000']

    def test_real_range_step_scale(self):
        rr = RealRange({'id': 'test', 'min': '0.001', 'max': '1000.0', 'scale': '10',
                        'step': 0.2})
        assert rr.to_list() == ['0.001', '0.210', '2.300', '23.200', '232.200']

    def test_real_range_exponent(self):
        rr = RealRange({'id': 'test', 'min': '1e-6', 'max': '1e3', 'scale': '1e1',
                        'step': 0.2})
        assert rr.to_list() == ['0.000001', '0.20001', '2.2001', '22.201', '222.21']

    def test_real_range_optional(self):
        rr = RealRange({'id': 'test', 'min': '0.001', 'max': '10.0', 'scale': '10',
                        'optional': True})
        assert rr.to_list() == [None, '0.001', '0.010', '0.100', '1.000', '10.000']

    def test_choice(self):
        cr = ChoiceRange({'id': 'test', 'values': ['One', 'Two', 'Three']})
        assert cr.to_list() == ['One', 'Two', 'Three']

    def test_choice_optional(self):
        cr = ChoiceRange({'id': 'test', 'values': ['One', 'Two', 'Three'],
                          'optional': 'true'})
        assert cr.to_list() == [None, 'One', 'Two', 'Three']

