#!/bin/bash

set -ef -o pipefail

# This script installs a new python 3.6 and then sets up a virtualenv for
# astrocyte dependencies.

# Where am I?
# http://stackoverflow.com/questions/59895/can-a-bash-script-tell-which-directory-it-is-stored-in

SOURCE="${BASH_SOURCE[0]}"
while [ -h "$SOURCE" ]; do # resolve $SOURCE until the file is no longer a symlink
  DIR="$( cd -P "$( dirname "$SOURCE" )" && pwd )"
  SOURCE="$(readlink "$SOURCE")"
  [[ $SOURCE != /* ]] && SOURCE="$DIR/$SOURCE" # if $SOURCE was a relative symlink, we need to resolve it relative to the path where the symlink file was located
done
DIR="$( cd -P "$( dirname "$SOURCE" )" && pwd )"

BASE_DIR="$DIR/.."
PYTHON_DIR="$BASE_DIR/.venv"
TMP_DIR="$PYTHON_DIR/src"

mkdir -p "$PYTHON_DIR"
mkdir -p "$TMP_DIR"


# Install OpenSSL from source
echo "Installing OpenSSL 1.1.1d"
cd "$TMP_DIR"

if [ ! -f "$PYTHON_DIR/bin/openssl" ]; then
    if [ ! -f "openssl-1.1.1d.tar.gz"  ]; then
        wget --quiet https://www.openssl.org/source/openssl-1.1.1d.tar.gz
    fi

    tar -zxf openssl-1.1.1d.tar.gz
    cd openssl-1.1.1d
    ./config --prefix="$PYTHON_DIR" --openssldir="$PYTHON_DIR/ssl"
    make -j --quiet
    make install > /dev/null 2>&1
fi

# used by python installation and pip
export LD_LIBRARY_PATH=$LD_LIBRARY_PATH:$PYTHON_DIR/lib
export PATH=$$PYTHON_DIR/bin/:$PATH

# Install SQLite from source
echo "Installing SQLite 3.31.1"
cd "$TMP_DIR"

if [ ! -x "$PYTHON_DIR/bin/sqlite3" ]; then
    if [ ! -f "sqlite-autoconf-3310100.tar.gz"  ]; then
        wget --quiet https://www.sqlite.org/2020/sqlite-autoconf-3310100.tar.gz
    fi

    tar -zxf sqlite-autoconf-3310100.tar.gz
    cd sqlite-autoconf-3310100
    ./configure --prefix="$PYTHON_DIR" --quiet
    make -j --quiet
    make install > /dev/null 2>&1
fi

# Install Python 3.6.10 from source with ssl supporting
echo "Installing Python 3.6.10"
cd "$TMP_DIR"

if [ ! -x "$PYTHON_DIR/bin/python" ]; then
    if [ ! -f "Python-3.6.10.tgz"  ]; then
        wget --quiet http://www.python.org/ftp/python/3.6.10/Python-3.6.10.tgz
    fi

    tar -zxf Python-3.6.10.tgz
    cd Python-3.6.10

    LD_RUN_PATH=$PYTHON_DIR/lib ./configure LDFLAGS="-L$PYTHON_DIR/lib" CPPFLAGS="-I$PYTHON_DIR/include" --prefix=$PYTHON_DIR --quiet
    LD_RUN_PATH=$PYTHON_DIR/lib make -j --quiet
    make install > /dev/null 2>&1
fi

# Install pip
echo "Installing pip"
cd "$TMP_DIR"

if [ ! -x "$PYTHON_DIR/bin/pip3" ]; then
    if [ ! -f "get-pip.py" ]; then
        wget --quiet https://bootstrap.pypa.io/get-pip.py
    fi

    "$PYTHON_DIR/bin/python" get-pip.py --no-cache-dir
fi

"$PYTHON_DIR/bin/pip3" install --upgrade pip

# Install basic requirements first
cd "$BASE_DIR"
"$PYTHON_DIR/bin/pip3" install --no-cache-dir numpy "setuptools>=41.0.0"

"$PYTHON_DIR/bin/pip3" install --no-cache-dir -r requirements.txt

"$PYTHON_DIR/bin/pip3" install --no-cache-dir -r requirements_dev.txt

# clean up
rm -fr "$TMP_DIR"

cd "$BASE_DIR"

echo
echo "Complete"
echo
